#!/usr/bin/env python3
# qautomome.py
# automome - the OTTish AUTOMEME clone
# Copyright (C) 2014  Penguin Development
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import csv
import random
import re
import os.path
import sys

from PySide import QtCore, QtGui, QtSvg

import automome

class SvgButton(QtSvg.QSvgWidget):
    """Extended SVG widget that maintains aspect ratio and has
 click signals. Aspect ratio code courtesy of Sean Fisk
 (https://github.com/seanfisk)."""

    forward = QtCore.Signal()
    back = QtCore.Signal()

    def paintEvent(self, paint_event):
        painter = QtGui.QPainter(self)
        default_width, default_height = \
            self.renderer().defaultSize().toTuple()
        widget_width, widget_height = self.size().toTuple()
        ratio_x = widget_width / default_width
        ratio_y = widget_height / default_height
        if ratio_x < ratio_y:
            new_width = widget_width
            new_height = widget_width * default_height / default_width
            new_left = 0
            new_top = (widget_height - new_height) / 2
        else:
            new_width = widget_height * default_width / default_height
            new_height = widget_height
            new_left = (widget_width - new_width) / 2
            new_top = 0
        self.renderer().render(painter, QtCore.QRectF(new_left,
            new_top, new_width, new_height))

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self.forward.emit()
        elif event.button() == QtCore.Qt.RightButton:
            self.back.emit()

    def keyPressEvent(self, event):
        if event.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return,
                QtCore.Qt.Key_Space, QtCore.Qt.Key_Right):
            if not event.modifiers() & QtCore.Qt.ShiftModifier:
                self.forward.emit()
            else:
                self.back.emit()
        elif event.key() in (QtCore.Qt.Key_Backspace, QtCore.Qt.Key_Left):
            if not event.modifiers() & QtCore.Qt.ShiftModifier:
                self.back.emit()
            else:
                self.forward.emit()
        else:
            QtSvg.QSvgWidget.keyPressEvent(self, event)

class QAMWindow(QtGui.QTabWidget):
    def __init__(self):
        QtGui.QTabWidget.__init__(self)
        self.setWindowTitle('AUTOMOME')

        self.settings = QtCore.QSettings('Penguin Development', 'qautomome')
        if not self.settings.contains('filter/expression'):
            self.settings.setValue('filter/expression', '.*')

        self.memequeue = []
        self.queuemax = int(self.settings.value('queue_max', 0))
        if self.queuemax < 1:
            self.queuemax = 1024
            self.settings.setValue('queue_max', 1024)
        self.filtertries = int(self.settings.value('filter/max_tries', 0))
        if self.filtertries < 1:
            self.filtertries = 1024
            self.settings.setValue('filter/max_tries', 1024)
        self.memeidx = -1

        self.xkcdfont = QtGui.QFontDatabase.addApplicationFont(
            'fonts/xkcd-Regular.otf')
        self.beanfont = QtGui.QFontDatabase.addApplicationFont(
            'fonts/beanish-Regular.otf')

        self.init_layout()
        self.nodisplay = False

        cats = str(self.settings.value('categories', ''))
        self.memes = []
        if 'a' in cats:
            self.memes += automome.memes_automeme_molpish
            self.cat_a_box.setChecked(True)
        if 'o' in cats:
            self.memes += automome.memes_other_molpish
            self.cat_o_box.setChecked(True)
        if 'm' in cats:
            self.memes += automome.memes_meta_molpish
            self.cat_m_box.setChecked(True)
        if 'A' in cats:
            self.memes += automome.memes_automeme_unmolpish
            self.cat_A_box.setChecked(True)
        if 'O' in cats:
            self.memes += automome.memes_other_unmolpish
            self.cat_O_box.setChecked(True)
        if 't' in cats:
            self.memes += automome.memes_ottomeme
            self.cat_t_box.setChecked(True)
        if 'F' in cats:
            self.memes += automome.memes_foreign
            self.cat_F_box.setChecked(True)

        if len(self.memes) == 0:
            self.memes = automome.memes_automeme_molpish + \
                automome.memes_other_molpish + \
                automome.memes_meta_molpish
            self.settings.setValue('categories', 'aom')

        self.setMinimumSize(448, 240)
        self.filter_changed()

        self.cuegan.forward.connect(self.do_meme)
        self.cuegan.back.connect(self.prev_meme)
        self.cat_a_box.stateChanged.connect(self.cats_changed)
        self.cat_o_box.stateChanged.connect(self.cats_changed)
        self.cat_m_box.stateChanged.connect(self.cats_changed)
        self.cat_A_box.stateChanged.connect(self.cats_changed)
        self.cat_O_box.stateChanged.connect(self.cats_changed)
        self.cat_t_box.stateChanged.connect(self.cats_changed)
        self.cat_F_box.stateChanged.connect(self.cats_changed)
        self.filterbox.editingFinished.connect(self.filter_changed)
        self.lockbox.stateChanged.connect(self.lock_changed)

        copy = lambda field: lambda: \
            QtGui.QApplication.clipboard().setText(field.toPlainText())

        self.copytextbutton.clicked.connect(copy(self.copytextbox))
        self.copybbcbutton.clicked.connect(copy(self.copybbcbox))
        self.copyhtmlbutton.clicked.connect(copy(self.copyhtmlbox))

        self.expgengroup.buttonClicked[int].connect(lambda i:
                self.exportspinner.setEnabled(i == 1))
        self.exportbutton.clicked.connect(self.export)

        self.cuegan.setFocus(QtCore.Qt.OtherFocusReason)

    def init_layout(self):
        ### Meme tab
        self.memetab = QtGui.QWidget()
        self.memelayout = QtGui.QGridLayout()
        self.memelayout.setSizeConstraint(QtGui.QLayout.SetMinAndMaxSize)

        self.cuegan = SvgButton('cuegan.svg')
        self.memelayout.addWidget(self.cuegan, 0, 0)
        self.cuegan.setMinimumSize(64, 128)
        self.cuegan.setToolTip('Click for new meme; '
            'right-click for previous meme')
        self.cuegan.setFocusPolicy(QtCore.Qt.StrongFocus)

        self.memelabel = QtGui.QLabel('CLICK CUEGAN, RECEIVE MEME')
        if self.xkcdfont != -1:
            self.memelabel.setStyleSheet('font-family: xkcd;')
        f = self.memelabel.font()
        f.setPointSizeF((2.0 if self.xkcdfont == -1 else 2.5) * \
            f.pointSizeF())
        self.memelabel.setFont(f)
        self.memelabel.setTextFormat(QtCore.Qt.RichText)
        self.memelabel.setWordWrap(True)
        self.memelabel.setMinimumWidth(384)

        self.memelayout.addWidget(self.memelabel, 0, 1)
        self.memelayout.setColumnStretch(0, 1)
        self.memelayout.setColumnStretch(1, 4)
        self.memetab.setLayout(self.memelayout)
        self.addTab(self.memetab, 'Memes')

        ### Category tab
        self.cattab = QtGui.QWidget()
        self.catlayout = QtGui.QVBoxLayout()

        self.cat_a_box = QtGui.QCheckBox('Molpish AUTOMEME')
        self.cat_o_box = QtGui.QCheckBox('Molpish other')
        self.cat_m_box = QtGui.QCheckBox('Molpish meta-memes')
        self.cat_A_box = QtGui.QCheckBox('Unmolpish AUTOMEME')
        self.cat_O_box = QtGui.QCheckBox('Unmolpish other')
        self.cat_t_box = QtGui.QCheckBox('xkcd memes from azule\'s OTToMeme')
        self.cat_t_layout = QtGui.QHBoxLayout()
        self.cat_t_link = QtGui.QLabel(
            '<a href="http://xkcd.modularthought.com/ottomeme/">(link)</a>')
        self.cat_t_link.setOpenExternalLinks(True)
        self.cat_F_box = QtGui.QCheckBox(
            'Foreign-language (may be unmolpish)')

        self.catlayout.addWidget(self.cat_a_box)
        self.catlayout.addWidget(self.cat_o_box)
        self.catlayout.addWidget(self.cat_m_box)
        self.catlayout.addWidget(self.cat_A_box)
        self.catlayout.addWidget(self.cat_O_box)
        self.cat_t_layout.addWidget(self.cat_t_box)
        self.cat_t_layout.addWidget(self.cat_t_link)
        self.cat_t_layout.addStretch(1)
        self.catlayout.addLayout(self.cat_t_layout)
        self.catlayout.addWidget(self.cat_F_box)
        self.catlayout.addStretch(1)
        self.cattab.setLayout(self.catlayout)
        self.addTab(self.cattab, 'Categories')

        ### Filter tab
        self.filtertab = QtGui.QWidget()
        self.filterlayout = QtGui.QVBoxLayout()
        self.filterlabel = QtGui.QLabel('Enter a <a href="https://'
            'docs.python.org/3/library/re.html#regular-expression-syntax">'
            'regular expression</a> to only show memes matching that '
            'expression. Simply leave the box empty to show all memes.')
        self.filterlabel.setOpenExternalLinks(True)
        self.filterlabel.setWordWrap(True)
        self.filterboxlabel = QtGui.QLabel('Filter expression:')
        self.filterbox = QtGui.QLineEdit(str(self.settings.value(
            'filter/expression', '.*')))
        self.lockbox = QtGui.QCheckBox('Lock meme template')

        self.filterlayout.addWidget(self.filterlabel)
        self.filterlayout.addSpacing(16)
        self.filterlayout.addWidget(self.filterboxlabel)
        self.filterlayout.addWidget(self.filterbox)
        self.filterlayout.addSpacing(16)
        self.filterlayout.addWidget(self.lockbox)
        self.filterlayout.addStretch(1)
        self.filtertab.setLayout(self.filterlayout)
        self.addTab(self.filtertab, 'Filter')

        ### Copy/export tab
        self.copytab = QtGui.QWidget()
        self.copylayout = QtGui.QGridLayout()
        self.copytextlabel = QtGui.QLabel('Plaintext:')
        self.copytextbox = QtGui.QPlainTextEdit()
        self.copytextbox.setReadOnly(True)
        self.copybbclabel = QtGui.QLabel('BBCode:')
        self.copybbcbox = QtGui.QPlainTextEdit()
        self.copybbcbox.setReadOnly(True)
        self.copyhtmllabel = QtGui.QLabel('(X)HTML:')
        self.copyhtmlbox = QtGui.QPlainTextEdit()
        self.copyhtmlbox.setReadOnly(True)

        h = int(QtGui.QFontMetrics(self.copytextbox.font()).lineSpacing() \
                * 2.2 + 8)
        self.copytextbox.setMinimumHeight(h)
        self.copybbcbox.setMinimumHeight(h)
        self.copyhtmlbox.setMinimumHeight(h)

        self.copytextbutton = QtGui.QPushButton(
            QtGui.QIcon.fromTheme('edit-copy'), 'Copy')
        self.copybbcbutton = QtGui.QPushButton(
            QtGui.QIcon.fromTheme('edit-copy'), 'Copy')
        self.copyhtmlbutton = QtGui.QPushButton(
            QtGui.QIcon.fromTheme('edit-copy'), 'Copy')

        self.exportlayout = QtGui.QHBoxLayout()
        self.expgenlayout = QtGui.QVBoxLayout()
        self.expgenlayout.setSpacing(0)
        self.expgengroup = QtGui.QButtonGroup()
        self.exportradio = QtGui.QRadioButton('&Export history')
        self.exportradio.setChecked(True)
        self.generateradio = QtGui.QRadioButton('&Generate new')
        self.expgengroup.addButton(self.exportradio, 0)
        self.expgengroup.addButton(self.generateradio, 1)
        self.exportspinner = QtGui.QSpinBox()
        self.exportspinner.setRange(1, self.queuemax)
        self.exportspinner.setEnabled(False)
        self.exporttypecombo = QtGui.QComboBox()
        self.exporttypecombo.addItem('Comma-separated values (CSV)', 'csv')
        self.exporttypecombo.addItem('fortune', 'fortune')
        self.exportbutton = QtGui.QPushButton('Export')

        self.copylayout.addWidget(self.copytextlabel, 0, 0)
        self.copylayout.addWidget(self.copytextbox, 1, 0)
        self.copylayout.addWidget(self.copytextbutton, 1, 1)
        self.copylayout.setRowStretch(1, 1)
        self.copylayout.addWidget(self.copybbclabel, 2, 0)
        self.copylayout.addWidget(self.copybbcbox, 3, 0)
        self.copylayout.addWidget(self.copybbcbutton, 3, 1)
        self.copylayout.setRowStretch(3, 1)
        self.copylayout.addWidget(self.copyhtmllabel, 4, 0)
        self.copylayout.addWidget(self.copyhtmlbox, 5, 0)
        self.copylayout.addWidget(self.copyhtmlbutton, 5, 1)
        self.copylayout.setRowStretch(5, 1)
        self.copylayout.setColumnStretch(0, 1)
        self.expgenlayout.addWidget(self.exportradio)
        self.expgenlayout.addWidget(self.generateradio)
        self.exportlayout.addLayout(self.expgenlayout)
        self.exportlayout.addWidget(self.exportspinner)
        self.exportlayout.addWidget(self.exporttypecombo)
        self.copylayout.addLayout(self.exportlayout, 6, 0)
        self.copylayout.addWidget(self.exportbutton, 6, 1)

        self.copytab.setLayout(self.copylayout)
        self.addTab(self.copytab, 'Copy/export')

        ### About tab
        self.abouttab = QtGui.QWidget()
        self.aboutlayout = QtGui.QVBoxLayout()
        self.aboutlabel = QtGui.QLabel(
            '<p>AUTOMOME version 4.3 - the OTTish AUTOMEME clone<br />'
            'Copyright (C) 2014-2015 Penguin Development<br />'
            'Licensed GPL-3+; see COPYING or '
            '<a href="http://www.gnu.org/licenses/">'
            'http://www.gnu.org/licenses/</a>.</p><p>'
            'Stand-alone <a href="http://qt-project.org/">Qt</a>/'
            '<a href="http://qt-project.org/wiki/PySide">PySide</a> '
            'clone of the <a href="http://automome.penguindevelopment.org/">'
            'web version</a>.</p>'
            '<p>Packaged Sun 13 Dec 17:56:40 UTC 2015</p>'
            '<p>Tip: right-click Cuegan for the previous meme. Use the right '
            'and left arrow keys for leopard-based operation.</p>')
        self.aboutlabel.setTextFormat(QtCore.Qt.RichText)
        self.aboutlabel.setWordWrap(True)
        self.aboutlinkslabel = QtGui.QLabel('Useful links:')
        self.aboutlinkslayout1 = QtGui.QHBoxLayout()
        self.aboutlinks1 = (
            QtGui.QLabel('<a href="http://www.penguindevelopment.org/">'
                'Penguin Development</a>'),
            QtGui.QLabel(
                '<a href="http://proj.penguindevelopment.org/automome/">'
                'Source code</a>'),
            QtGui.QLabel('<a href="http://bugs.penguindevelopment.org/">'
                'Bugs</a>'))
        self.aboutlinkslayout2 = QtGui.QHBoxLayout()
        self.aboutlinks2 = (
            QtGui.QLabel('<a href="http://www.xkcd.com/">'
                'xkcd</a>'),
            QtGui.QLabel('<a href="http://xkcd.mscha.org/viewer/1">OTC</a>'),
            QtGui.QLabel(
                '<a href="http://forums.xkcd.com/viewtopic.php?f=7&t=101043">'
                'OTT</a>'),
            QtGui.QLabel(
                '<a href="http://xkcd-time.wikia.com/wiki/XKCD_Time_Wiki">'
                'OTW</a>'))

        self.aboutlabel.setOpenExternalLinks(True)
        self.aboutlayout.addWidget(self.aboutlabel)
        self.aboutlayout.addStretch(1)
        self.aboutlayout.addWidget(self.aboutlinkslabel)
        for l in self.aboutlinks1:
            l.setOpenExternalLinks(True)
            l.setWordWrap(True)
            l.setAlignment(QtCore.Qt.AlignHCenter)
            self.aboutlinkslayout1.addWidget(l)
        self.aboutlayout.addLayout(self.aboutlinkslayout1)
        for l in self.aboutlinks2:
            l.setOpenExternalLinks(True)
            l.setWordWrap(True)
            l.setAlignment(QtCore.Qt.AlignHCenter)
            self.aboutlinkslayout2.addWidget(l)
        self.aboutlayout.addLayout(self.aboutlinkslayout2)
        self.abouttab.setLayout(self.aboutlayout)
        self.addTab(self.abouttab, 'About')

        s = self.sizeHint()
        if s.isValid():
            self.resize(s)

    def display_meme(self):
        self.memelabel.setText(self.memequeue[self.memeidx][1])
        self.copytextbox.setPlainText(self.memequeue[self.memeidx][2])
        self.copybbcbox.setPlainText(self.memequeue[self.memeidx][3])
        self.copyhtmlbox.setPlainText(self.memequeue[self.memeidx][4])
        s = self.sizeHint()
        if self.width() > s.width():
            s.setWidth(self.width())
        if self.height() > s.height():
            s.setHeight(self.height())
        self.resize(s)

    @staticmethod
    def escape_entities(x):
        return x.replace('&', '&amp;').replace('<', '&lt;').replace('>',
            '&gt;')

    @staticmethod
    def unescape_entities(x):
        return x.replace('&gt;', '>').replace('&lt;', '<').replace(
            '&amp;', '&').replace('%SCS%','').replace('%SCE%', '')

    def beanify(self, x):
        return (re.sub(r'([᐀-ᙿᢰ-ᣵ]+)', lambda pat: \
                '<span style="font-family: beanish;">{}</span>'.format(
                    pat.group(1)), x) if self.beanfont != -1 else x)

    @staticmethod
    def smallcaps(x, bbc = False, fs = '85%'):
        return re.sub(r'%SCS%([a-z]+)%SCE%', lambda pat: \
            ('[size=85]{}[/size]' if bbc \
                else ('<span style="font-size: ' + fs + ';">{}</span>')
            ).format(pat.group(1).upper()), x)

    @QtCore.Slot()
    def do_meme(self):
        if self.memeidx == len(self.memequeue) - 1:
            success = False
            for t in range(self.filtertries):
                meme = self.memequeue[self.memeidx][0] \
                    if self.lockbox.isChecked() \
                    else random.choice(self.memes)
                m = rawtext = str(meme)
                if not self.filterexpr.search(m):
                    continue
                else:
                    success = True
                m = self.escape_entities(re.sub(r'([a-z]+)',
                    r'%SCS%\g<1>%SCE%', m))
                s = ''
                em_open = 0
                for i in range(len(m)):
                    if m[i] == '_':
                        if (i == 0 or m[i-1] in ' *') and i != len(m) - 1 \
                                and em_open == 0:
                            s += '<em>'
                            em_open = 1
                        elif (i == len(m) - 1 or m[i+1] in ',.;:?!* ') \
                                and em_open == 1:
                            s += '</em>'
                            em_open = 0
                        else:
                            s += m[i]
                    else:
                        s += m[i]

                if em_open != 0:
                    s += '</em>'

                self.memequeue.append((meme,
                    self.beanify(self.smallcaps(s, fs='small')), rawtext,
                    '[quote="[url=http://automome.penguindevelopment.org/]'
                        'AUTOMOME[/url]"]{}[/quote]'.format(
                            self.unescape_entities(self.smallcaps(
                                re.sub(r'<(/?)em>', r'[\g<1>i]', s), True))),
                    '<a href="http://automome.penguindevelopment.org/">'
                        '{}</a>'.format(self.smallcaps(s))))
                if success:
                    break

            if not success:
                QtGui.QMessageBox.information(self,
                    'No match found',
                    'No match found for filter expression '
                        '`{}\' after {} tries.'.format(self.filterbox.text(),
                            self.filtertries))
                return

        if len(self.memequeue) > self.queuemax:
            del self.memequeue[0]
        else:
            self.memeidx += 1
        if not self.nodisplay:
            self.display_meme()

    @QtCore.Slot()
    def prev_meme(self):
        if self.memeidx >= 1:
            self.memeidx -= 1
            self.display_meme()

    @QtCore.Slot(int)
    def cats_changed(self, null):
        cats = ''
        self.memes = []
        if self.cat_a_box.isChecked():
            cats += 'a'
            self.memes += automome.memes_automeme_molpish
        if self.cat_o_box.isChecked():
            cats += 'o'
            self.memes += automome.memes_other_molpish
        if self.cat_m_box.isChecked():
            cats += 'm'
            self.memes += automome.memes_meta_molpish
        if self.cat_A_box.isChecked():
            cats += 'A'
            self.memes += automome.memes_automeme_unmolpish
        if self.cat_O_box.isChecked():
            cats += 'O'
            self.memes += automome.memes_other_unmolpish
        if self.cat_t_box.isChecked():
            cats += 't'
            self.memes += automome.memes_ottomeme
        if self.cat_F_box.isChecked():
            cats += 'F'
            self.memes += automome.memes_foreign

        self.settings.setValue('categories', cats)

    @QtCore.Slot(int)
    def lock_changed(self, state):
        if len(self.memequeue) == 0 and state == QtCore.Qt.Checked:
            self.lockbox.setCheckState(QtCore.Qt.Unchecked)
            QtGui.QMessageBox.warning(self, 'No meme found',
                'You must generate a meme before locking the template.')
        else:
            del self.memequeue[self.memeidx + 1:]

    @QtCore.Slot()
    def filter_changed(self):
        try:
            self.filterexpr = re.compile(self.filterbox.text())
            self.settings.setValue('filter/expression', self.filterbox.text())
        except:
            text = self.filterbox.text()
            self.filterbox.setText(self.settings.value(
                'filter/expression', '.*'))
            QtGui.QMessageBox.warning(self,
                'Syntax error',
                'Invalid regular expression: {}'.format(text))

    @QtCore.Slot()
    def export(self):
        mode = self.expgengroup.checkedId()
        fname = os.path.realpath(QtGui.QFileDialog.getSaveFileName(self,
            'Export memes')[0])

        if mode == 1:
            queue = self.memequeue
            idx = self.memeidx
            self.nodisplay = True
            self.memequeue = []
            self.memeidx = -1
            for i in range(self.exportspinner.value()):
                self.do_meme()

        ftype = self.exporttypecombo.itemData(
                self.exporttypecombo.currentIndex())

        with open(fname, 'w') as f:
            if ftype == 'fortune':
                f.write('\n%\n'.join([x[2] for x in self.memequeue]))
            else:
                writer = csv.writer(f)
                for m in self.memequeue:
                    writer.writerow(m[2:])

        if mode == 1:
            self.nodisplay = False
            self.memequeue = queue
            self.memeidx = idx

if __name__ == '__main__':
    random.seed()

    app = QtGui.QApplication(sys.argv)
    win = QAMWindow()
    win.show()

    sys.exit(app.exec_())
